@@ -1,7 +1,5 @@
 /*****************************************************************************\
- *  $Id: hostlist.c 9950 2010-08-31 21:57:35Z grondo $
- *****************************************************************************
- *  $LSDId: hostlist.c 9950 2010-08-31 21:57:35Z grondo $
+ *  $LSDId: hostlist.c 11882 2012-10-03 17:31:41Z grondo $
  *****************************************************************************
  *  Copyright (C) 2002 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -353,8 +351,8 @@ static void _error(char *file, int line, char *msg, ...)
 }
 
 
-/*
- * Helper function for host list string parsing routines
+/* 
+ * Helper function for host list string parsing routines 
  * Returns a pointer to the next token; additionally advance *str
  * to the next separator.
  *
@@ -380,13 +378,13 @@ static char * _next_tok(char *sep, char **str)
     /* assign token ptr */
     tok = *str;
 
-    while ( **str != '\0' &&
+    while ( **str != '\0' && 
 	    (level != 0 || strchr(sep, **str) == NULL) ) {
       if ( **str == '[' ) level++;
       else if ( **str == ']' ) level--;
       (*str)++;
     }
-
+    
    /* nullify consecutive separators and push str beyond them */
     while (**str != '\0' && strchr(sep, **str) != '\0')
         *(*str)++ = '\0';
@@ -970,9 +968,13 @@ static int hostrange_hn_within(hostrange_t hr, hostname_t hn)
 
     /*
      *  Finally, check whether [hn], with a valid numeric suffix,
-     *   falls within the range of [hr].
+     *   falls within the range of [hr] if [hn] and [hr] prefix are
+     *   identical.
      */
-    if (hn->num <= hr->hi && hn->num >= hr->lo) {
+    if ((len_hr == len_hn)
+        && (strcmp (hn->prefix, hr->prefix) == 0)
+        && (hn->num <= hr->hi)
+        && (hn->num >= hr->lo)) {
         int width = hostname_suffix_width (hn);
         if (!_width_equiv(hr->lo, &hr->width, hn->num, &width))
             return -1;
