@@ -1,5 +1,5 @@
 /*****************************************************************************\
- *  $Id: mod.c 1227 2010-09-03 21:57:36Z grondo $
+ *  $Id$
  *****************************************************************************
  *  Copyright (C) 2001-2006 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -100,13 +100,13 @@ static int  _mod_load_static(int);
 #else
 static int  _mod_load_dynamic_modules(const char *, opt_t *);
 static int  _mod_load_dynamic(const char *);
-static int  _mod_initialize(mod_t mod);
-static int  _mod_init_list_safe(mod_t mod, void *arg);
 static int  _cmp_filenames(mod_t, char *);
 static int  _is_loaded(char *filename); 
 static bool _path_permissions_ok(const char *dir, uid_t pdsh_owner);
 static perm_error_t  _dir_permission_error(struct stat *, uid_t alt_uid);
 #endif
+static int  _mod_initialize(mod_t mod);
+static int  _mod_init_list_safe(mod_t mod, void *arg);
 static void _mod_destroy(mod_t mod);
 static bool _mod_opts_ok(mod_t mod);
 static int  _mod_print_info(mod_t mod);
@@ -175,6 +175,22 @@ _mod_postop(mod_t mod, opt_t *pdsh_opts)
 }
 
 
+/*
+ *  Like list_next (i), but skip over inactive modules
+ */
+static mod_t _mod_next_active (ListIterator i)
+{
+    mod_t mod;
+
+    while ((mod = list_next (i))) {
+        if (mod->initialized)
+            return (mod);
+    }
+
+    return (NULL);
+}
+
+
 /*
  *  Call any "read wcoll" functions exported by modules. The module
  *    is responsible for deciding when to generate a new wcoll, 
@@ -200,7 +216,7 @@ mod_read_wcoll(opt_t *opt)
         return -1;
     }
 
-    while ((mod = list_next(module_itr))) {
+    while ((mod = _mod_next_active (module_itr))) {
         hostlist_t hl = NULL;
 
         if (!(hl = _mod_read_wcoll(mod, opt)))
@@ -233,7 +249,7 @@ mod_postop(opt_t *pdsh_opts)
         return 1;
     }
 
-    while ((mod = list_next(module_itr)))
+    while ((mod = _mod_next_active (module_itr)))
         errors += _mod_postop(mod, pdsh_opts);
 
     list_iterator_destroy(module_itr);
@@ -342,20 +358,6 @@ static int _mod_initialize_modules_by_name (char *names, List m)
     return (0);
 }
 
-static void _mod_delete_uninitialized (List l)
-{
-    mod_t mod;
-    ListIterator i = list_iterator_create (l);
-
-    while ((mod = list_next (i))) {
-        if (!mod->initialized)
-            list_delete (i);
-    }
-
-    list_iterator_destroy (i);
-}
-
-
 int mod_load_modules(const char *dir, opt_t *opt)
 {
     int rc = 0;
@@ -377,11 +379,6 @@ int mod_load_modules(const char *dir, opt_t *opt)
      */
     list_for_each (module_list, (ListForF) _mod_init_list_safe, NULL);
 
-    /*
-     *  Remove all uninitialized modules
-     */
-    _mod_delete_uninitialized (module_list);
-
     return(rc);
 }
 
@@ -434,6 +431,7 @@ _mod_print_info(mod_t mod)
     out("Module: %s/%s\n",    mod->pmod->type, mod->pmod->name); 
     out("Author: %s\n",       mod->pmod->author ? mod->pmod->author : "???");
     out("Descr:  %s\n",       mod->pmod->descr ? mod->pmod->descr : "???");
+    out("Active: %s\n",       mod->initialized ? "yes" : "no");
 
     if (mod->pmod->opt_table && mod->pmod->opt_table->opt) {
         out("Options:\n");
@@ -489,8 +487,8 @@ mod_count(char *type)
     return i;
 }
 
-List
-mod_get_module_names(char *type)
+static List
+_mod_get_module_names(char *type, int get_active)
 {
     List l;
     mod_t mod;
@@ -506,15 +504,14 @@ mod_get_module_names(char *type)
         return NULL;
     }
 
-    if (type == NULL) {
-        while((mod = list_next(module_itr)))
+    while ((mod = (type ? list_find(module_itr, (ListFindF) _cmp_type, type)
+                        : list_next(module_itr)))) {
+        /*
+         *  Push active (initialized) modules onto list if get_active
+         *   is true, otherwise push inactive (!initialized) modules:
+         */
+        if (!get_active == !mod->initialized)
             list_push(l, mod->pmod->name);
-        list_iterator_destroy(module_itr);
-        return l;
-    }
-
-    while ((mod = list_find(module_itr, (ListFindF) _cmp_type, type))) {
-        list_push(l, mod->pmod->name);
     }
 
     list_iterator_destroy(module_itr);
@@ -522,6 +519,17 @@ mod_get_module_names(char *type)
     return l;
 }
 
+List mod_get_module_names (char *type)
+{
+    return _mod_get_module_names (type, 1);
+}
+
+List mod_get_uninitialized_module_names (char *type)
+{
+    return _mod_get_module_names (type, 0);
+}
+
+
 void
 mod_list_module_info(void)
 {
@@ -648,7 +656,7 @@ mod_process_opt(opt_t *opt, int c, char *optarg)
         return -1;
     }
 
-    while ((mod = list_next(module_itr))) {
+    while ((mod = _mod_next_active (module_itr))) {
         if ((p = _mod_find_opt(mod, c))) {
             list_iterator_destroy(module_itr);
             return p->f(opt, c, optarg);
@@ -879,7 +887,7 @@ _mod_load_dynamic_modules(const char *dir, opt_t *pdsh_opts)
     if (!initialized) 
         mod_init();
 
-    if (_pdsh_owner(pdsh_opts->path_progname, &pdsh_owner) < 0)
+    if (_pdsh_owner(pdsh_opts->local_program_path, &pdsh_owner) < 0)
         return -1;
 
     if (!_path_permissions_ok(dir, pdsh_owner)) 
