@@ -1,5 +1,5 @@
 /*****************************************************************************\
- *  $Id: opt.c 1226 2010-09-03 21:57:35Z grondo $
+ *  $Id$
  *****************************************************************************
  *  Copyright (C) 2001-2006 The Regents of the University of California.
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
@@ -43,6 +43,9 @@
 
 #include <errno.h>
 
+#include <regex.h>
+#include <ctype.h>
+
 #include "src/common/hostlist.h"
 #include "src/common/err.h"
 #include "src/common/list.h"
@@ -55,11 +58,11 @@
 #include "mod.h"
 #include "rcmd.h"
 
-static void _usage(opt_t * opt);
-static void _show_version(void);
-
-static char *exclude_buf = NULL;
-
+/*
+ *  Fallback maximum username length if sysconf(_SC_LOGIN_NAME_MAX) not
+ *   available or fails. (buffers will be this size + 1 for NUL termination)
+ */
+#define DEFAULT_MAX_USERNAME_LENGTH 16
 
 #define OPT_USAGE_DSH "\
 Usage: pdsh [-options] command ...\n\
@@ -117,6 +120,7 @@ Usage: rpdcp [-options] src [src2...] dir\n\
 #define PCP_ARGS	"pryzZe:"
 #define GEN_ARGS	"hLNKR:M:t:cqf:w:x:l:u:bI:dVT:Q"
 
+
 /*
  *  Pdsh options string (for getopt) -- initialized
  *    in _init_pdsh_options(), and appended to by modules that
@@ -145,7 +149,7 @@ static int    remote_argc;
 
 const char **pdsh_remote_argv (void)
 {
-    return remote_argv;
+    return (const char **) remote_argv;
 }
 
 int pdsh_remote_argc (void)
@@ -153,6 +157,19 @@ int pdsh_remote_argc (void)
     return remote_argc;
 }
 
+/*
+ *  List of explicitly excluded hosts and regex filter options:
+ */
+static List exclude_list = NULL;
+static List regex_list = NULL;
+
+static void _usage(opt_t * opt);
+static void _show_version(void);
+static int wcoll_args_process (opt_t *opt, char *args);
+static void wcoll_apply_regex (opt_t *opt, List regexs);
+static void wcoll_apply_excluded (opt_t *opt, List excludes);
+static void wcoll_expand (opt_t *opt);
+
 static void
 _init_pdsh_options()
 {
@@ -278,6 +295,40 @@ done:
     return abspath;
 }
 
+
+/*
+ *   Return a value for the max login name length (username buffer length)
+ */
+static int login_name_max_len (void)
+{
+    static int maxnamelen = -1;
+
+    if (maxnamelen < 0) {
+#ifdef _SC_LOGIN_NAME_MAX
+        errno = 0;
+        if ((maxnamelen = sysconf (_SC_LOGIN_NAME_MAX)) <= 0) {
+            err ("%p: sysconf(LOGIN_NAME_MAX): %m\n");
+            maxnamelen = DEFAULT_MAX_USERNAME_LENGTH;
+        }
+#else
+        maxnamelen = DEFAULT_MAX_USERNAME_LENGTH;
+#endif
+    }
+
+    return (maxnamelen);
+}
+
+static void copy_username (char *dst, const char *src)
+{
+    int maxlen = login_name_max_len ();
+
+    if (strlen (src) > maxlen)
+        errx ("%p: Fatal: username '%s' exceeds max username length (%d)\n",
+                src, maxlen);
+
+    strcpy (dst, src);
+}
+
 /*
  * Set defaults for various options.
  *	opt (IN/OUT)	option struct
@@ -287,6 +338,8 @@ void opt_default(opt_t * opt, char *argv0)
     struct passwd *pw;
 
     opt->progname = xbasename(argv0);
+    opt->luser = Malloc (login_name_max_len () + 1);
+    opt->ruser = Malloc (login_name_max_len () + 1);
 
     opt->reverse_copy = false;
 
@@ -306,10 +359,8 @@ void opt_default(opt_t * opt, char *argv0)
         _init_pdsh_options();
 
     if ((pw = getpwuid(getuid())) != NULL) {
-        strncpy(opt->luser, pw->pw_name, MAX_USERNAME);
-        strncpy(opt->ruser, pw->pw_name, MAX_USERNAME);
-        opt->luser[MAX_USERNAME - 1] = '\0';
-        opt->ruser[MAX_USERNAME - 1] = '\0';
+        copy_username (opt->luser, pw->pw_name);
+        copy_username (opt->ruser, pw->pw_name);
         opt->luid = pw->pw_uid;
     } else
         errx("%p: who are you?\n");
@@ -351,7 +402,13 @@ void opt_default(opt_t * opt, char *argv0)
     opt->separate_stderr = true;
 #endif
 
-    opt->path_progname = _find_path(argv0);
+    opt->local_program_path = _find_path(argv0);
+
+    /*
+     *  By default assume remote path to pdsh/pdcp is the same
+     *   as local path: (overridden with pdpc -e or PDSH_REMOTE_PDCP_PATH).
+     */
+    opt->remote_program_path = Strdup(opt->local_program_path);
 
     /* PCP specific */
     opt->outfile_name = NULL;
@@ -416,9 +473,15 @@ void opt_env(opt_t * opt)
             xstrcat(&opt->dshpath, ";");
         }
     }
+
+    if (pdsh_personality() == PCP) {
+        if ((rhs = getenv ("PDSH_REMOTE_PDCP_PATH")) != NULL) {
+            Free ((void **) &opt->remote_program_path);
+            opt->remote_program_path = Strdup (rhs);
+        }
+    }
 }
 
-static void wcoll_append (opt_t *opt, char *hosts);
 
 /*
  *  Process any options that need to be handled early, i.e.
@@ -453,6 +516,24 @@ void opt_args_early (opt_t * opt, int argc, char *argv[])
     }
 }
 
+static void wcoll_append_excluded (opt_t *opt, char *exclude_args)
+{
+    List l = list_split (",", exclude_args);
+    ListIterator i = list_iterator_create (l);
+    char *s;
+
+    while ((s = list_next (i))) {
+        char *p = NULL;
+        xstrcatchar (&p, '-');
+        xstrcat (&p, s);
+        wcoll_args_process (opt, p);
+        Free ((void **) &p);
+    }
+
+    list_iterator_destroy (i);
+    list_destroy (l);
+}
+
 /*
  * Override  default/environment options with command line arguments.
  *	opt (IN/OUT)	option struct	
@@ -509,12 +590,12 @@ void opt_args(opt_t * opt, int argc, char *argv[])
             break;
         case 'w':              /* target node list */
             if (strcmp(optarg, "-") == 0)
-                opt->wcoll = read_wcoll(NULL, stdin);
-            else 
-                wcoll_append(opt, optarg);
+                wcoll_args_process (opt, "^-");
+            else
+                wcoll_args_process (opt, optarg);
             break;
         case 'x':              /* exclude node list */
-            exclude_buf = Strdup(optarg);
+            wcoll_append_excluded (opt, optarg);
             break;
         case 'q':              /* display fanout and wcoll then quit */
             opt->info_only = true;
@@ -535,8 +616,7 @@ void opt_args(opt_t * opt, int argc, char *argv[])
             opt->sigint_terminates = true;
             break;
         case 'l':              /* specify remote username for rshd */
-            strncpy(opt->ruser, optarg, MAX_USERNAME);
-            opt->ruser[MAX_USERNAME - 1] = '\0';
+            copy_username (opt->ruser, optarg);
             break;
         case 'r':              /* rcp: copy recursively */
             if (pdsh_personality() == PCP)
@@ -552,9 +632,9 @@ void opt_args(opt_t * opt, int argc, char *argv[])
             break;
         case 'e':
             if (pdsh_personality() == PCP) {
-                Free ((void **) &opt->path_progname);
-                opt->path_progname = Strdup(optarg);
-            } 
+                Free ((void **) &opt->remote_program_path);
+                opt->remote_program_path = Strdup(optarg);
+            }
             else
                 goto test_module_option;
             break;
@@ -642,27 +722,64 @@ void opt_args(opt_t * opt, int argc, char *argv[])
         }
     }
 
-    /* ignore wcoll & -x option if we are running pcp server */
-    if (opt->pcp_server) 
+    /* ignore wcoll filtering when running pcp server */
+    if (opt->pcp_server)
         return;
 
     /*
-     *  Attempt to grab wcoll from modules stack unless
-     *    wcoll was set from -w.
+     *  Give modules a chance to fill in wcoll if it hasn't been already:
      */
-    if (mod_read_wcoll(opt) < 0)
-        exit(1);
+    if (mod_read_wcoll (opt) < 0)
+        exit (1);
 
     /*
-     *  If wcoll is still empty, try WCOLL environment variable.
+     *  If wcoll is still empty, try WCOLL env variable:
      */
-    if (!opt->wcoll) {
-        char *val;
-        if ((val = getenv("WCOLL")) != NULL)
-            opt->wcoll = read_wcoll(val, NULL);
+    if (opt->wcoll == NULL) {
+        char *val = getenv ("WCOLL");
+        if (val != NULL)
+            opt->wcoll = read_wcoll (val, NULL);
+    }
+
+    if (opt->wcoll) {
+        /*
+         *  Now apply wcoll filtering
+         */
+        if (exclude_list) {
+            wcoll_apply_excluded (opt, exclude_list);
+            list_destroy (exclude_list);
+        }
+        if (regex_list) {
+            wcoll_apply_regex (opt, regex_list);
+            list_destroy (regex_list);
+        }
+
+        /*
+         *  Finally, re-expand wcoll to allow two sets of brackets.
+         *   (For historical compatibility)
+         */
+        wcoll_expand (opt);
     }
 }
 
+static void wcoll_expand (opt_t *opt)
+{
+    hostlist_t hl = opt->wcoll;
+    const char *hosts;
+
+    /*
+     *  Create new hostlist for wcoll
+     */
+    opt->wcoll = hostlist_create ("");
+    while ((hosts = hostlist_shift (hl))) {
+        hostlist_push (opt->wcoll, hosts);
+        free (hosts);
+    }
+
+    hostlist_destroy (hl);
+}
+
+
 /* 
  * Check if infile_names are legit.
  */
@@ -710,21 +827,6 @@ bool opt_verify(opt_t * opt)
     if (mod_postop(opt) > 0)
         verified = false;
 
-    /* handle -x option */
-    if (!opt->pcp_server && exclude_buf && opt->wcoll) {
-        /*
-         * Delete any hosts in exclude_buf from the wcoll,  
-         *  ignoring errors (except for an invalid hostlist
-         *  in exclude_buf)
-         */
-        if (hostlist_delete(opt->wcoll, exclude_buf) <= 0) {
-            if (errno == EINVAL)
-                errx ("%p: Invalid argument to -x \"%s\"\n", exclude_buf);
-        }
-
-        Free((void **) &exclude_buf);
-    }
-
     /* can't prompt for command if stdin was used for wcoll */
     if (personality == DSH && opt->stdin_unavailable && !opt->cmd) {
         _usage(opt);
@@ -874,7 +976,6 @@ void opt_list(opt_t * opt)
         out("Outfile			%s\n", STRORNULL(opt->outfile_name));
         out("Recursive		%s\n", BOOLSTR(opt->recursive));
         out("Preserve mod time/mode	%s\n", BOOLSTR(opt->preserve));
-        out("Full program pathname	%s\n", STRORNULL(opt->path_progname));
         if (opt->pcp_server) {
             out("pcp server         	%s\n", BOOLSTR(opt->pcp_server));
             out("target is directory	%s\n", BOOLSTR(opt->target_is_directory));
@@ -882,7 +983,9 @@ void opt_list(opt_t * opt)
     }
 
     if (!opt->pcp_server) {
-
+        out("Full program pathname	%s\n", STRORNULL(opt->local_program_path));
+        out("Remote program path	%s\n",
+                STRORNULL(opt->remote_program_path));
         out("\n-- Generic options --\n");
         out("Local username		%s\n", opt->luser);
         out("Local uid     		%d\n", opt->luid);
@@ -929,10 +1032,16 @@ void opt_free(opt_t * opt)
         Free((void **) &pdsh_options);
     if (opt->dshpath)
         Free((void **) &opt->dshpath);
-    if (opt->path_progname)
-        Free((void **) &opt->path_progname);
+    if (opt->local_program_path)
+        Free((void **) &opt->local_program_path);
+    if (opt->remote_program_path)
+        Free((void **) &opt->remote_program_path);
     if (opt->infile_names)
         list_destroy(opt->infile_names);
+    if (opt->luser)
+        Free((void **) &opt->luser);
+    if (opt->ruser)
+        Free((void **) &opt->ruser);
 
     rcmd_exit();
 }
@@ -956,6 +1065,22 @@ static int _module_list_string(char *type, char *buf, int len)
     return (n);
 }
 
+static int _module_list_uninitialized (char *type, char *buf, int len)
+{
+    List l = NULL;
+    int  n = 0;
+
+    if (mod_count(type) == 0)
+        return (0);
+
+    l = mod_get_uninitialized_module_names(type);
+    n = list_join(buf, len, ",", l);
+    list_destroy(l);
+
+    return (n);
+
+}
+
 static char *_rcmd_module_list(char *buf, int maxlen)
 {
     int len, len2;
@@ -1024,7 +1149,15 @@ static void _show_version(void)
     printf("rcmd modules: %s\n", _rcmd_module_list(buf, sizeof (buf)));
 
     n = _module_list_string("misc", buf, sizeof (buf));
-    printf("misc modules: %s\n", n ? buf : "(none)");
+    printf("misc modules: %s", n ? buf : "(none)");
+
+    if ((n = _module_list_uninitialized ("misc", buf, sizeof (buf)))) {
+        printf (" (*conflicting: %s)\n", buf);
+	printf ("[* To force-load a conflicting module,"
+	        " use the -M <name> option]\n");
+    }
+    else
+        printf ("\n");
 
     exit(0);
 }
@@ -1072,39 +1205,207 @@ static int get_host_rcmd_type (char *hosts, char **rptr, char **hptr,
     }
 
     return (1);
-    
+}
+
+void free_f (void *x)
+{
+    Free (&x);
+}
+
+struct regex_info {
+    int     exclude;
+    int     cflags;
+    int     eflags;
+    int     compiled;
+    char *  pattern;
+    regex_t reg;
+};
+
+void regex_info_destroy (struct regex_info *re)
+{
+    if (re->compiled)
+        regfree (&re->reg);
+    if (re->pattern)
+        Free ((void **) &re->pattern);
+    Free ((void **) &re);
+}
+
+struct regex_info * regex_info_create (const char *r, int exclude)
+{
+    int rc;
+    struct regex_info *re = Malloc (sizeof (*re));
+
+    re->pattern = Strdup (r);
+    re->exclude = exclude;
+    re->cflags =  REG_EXTENDED | REG_NOSUB;
+    re->eflags =  0;
+    re->compiled = 0;
+
+    if ((rc = regcomp (&re->reg, re->pattern, re->cflags)) != 0) {
+        char msg [4096];
+        regerror (rc, &re->reg, msg, sizeof (msg));
+        err ("%p: Error %s pattern \"%s\": %s\n",
+                re->exclude ? "excluding" : "matching",
+                re->pattern, msg);
+        regex_info_destroy (re);
+        return (NULL);
+    }
+
+    re->compiled = 1;
+    return (re);
 }
 
 
-static void wcoll_append (opt_t *opt, char *str)
+void hostlist_filter_regex (hostlist_t hl, struct regex_info *re)
 {
-    hostlist_t  hl = hostlist_create (str);
-    char *rcmd_type = NULL;
-    char *tok;
+    char *host;
+    hostlist_iterator_t i;
 
-    if (!hl)
-        return;
+    i = hostlist_iterator_create (hl);
+    while ((host = hostlist_next (i))) {
+        int rc = regexec (&re->reg, host, 0, NULL, re->eflags);
+        if ((re->exclude && rc == 0) || (!re->exclude && rc == REG_NOMATCH))
+            hostlist_remove (i);
+        free (host);
+    }
+    hostlist_iterator_destroy (i);
+}
 
-    if (!opt->wcoll)
-        opt->wcoll = hostlist_create (NULL);
 
-    while ((tok = hostlist_shift (hl))) {
-        char *hosts, *user = NULL;
+static void list_push_hostlist (List l, hostlist_t hl)
+{
+    size_t n = 4096;
+    char *s = Malloc (n);
 
-        get_host_rcmd_type (tok, &rcmd_type, &hosts, &user);
+    while ((hostlist_ranged_string (hl, n-1, s) < 0) && (n*=2 < 0x7fffff)) {
+        Realloc ((void **) &s, n);
+    }
 
-        hostlist_push (opt->wcoll, hosts);
+    list_push (l, s);
+}
 
-        if (rcmd_type || user) {
-            if (rcmd_register_defaults (hosts, rcmd_type, user) < 0)
-                errx ("Failed to register rcmd module \"%s\" for hosts \"%s\"\n",
-                      rcmd_type, hosts);
+
+static void hostlist_assign (hostlist_t *hlp, hostlist_t hl2)
+{
+    if (*hlp == NULL)
+        *hlp = hostlist_create ("");
+    hostlist_push_list (*hlp, hl2);
+}
+
+static int wcoll_arg_process (char *arg, opt_t *opt)
+{
+    struct regex_info *re;
+    int excluded = 0;
+    char *p = arg;
+
+    if (exclude_list == NULL)
+        exclude_list = list_create (free_f);
+
+    if (regex_list == NULL)
+        regex_list = list_create ((ListDelF) regex_info_destroy);
+
+    /*
+     *  Check for excluded arg
+     */
+    if (*p == '-') {
+        excluded = 1;
+        p++;
+    }
+
+    /*
+     *  Move past any leading whitespace
+     */
+    while (isspace (*p))
+        p++;
+
+    if (*p == '^') {
+        hostlist_t hl = read_wcoll (p+1, NULL);
+        if (hl == NULL)
+            errx ("%p: Error reading wcoll: %s: %m\n", p+1);
+        if (excluded)
+            list_push_hostlist (exclude_list, hl);
+        else
+            hostlist_assign (&opt->wcoll, hl);
+        hostlist_destroy (hl);
+    }
+    else if (*p == '/') {
+        int len;
+
+        ++p;
+        len = strlen (p);
+        if (p [len - 1] == '/')
+            p [len - 1] = '\0';
+
+        if ((re = regex_info_create (p, excluded)) == NULL)
+            errx ("%p: Fatal error\n");
+
+        list_push (regex_list, re);
+    }
+    else {
+        if (excluded) {
+            list_push (exclude_list, Strdup (p));
+        }
+        else {
+            char *rcmd_type = NULL;
+            char *hosts, *user = NULL;
+
+            if (!opt->wcoll)
+                opt->wcoll = hostlist_create ("");
+
+            get_host_rcmd_type (p, &rcmd_type, &hosts, &user);
+            hostlist_push (opt->wcoll, hosts);
+            if (rcmd_type || user) {
+                if (rcmd_register_defaults (hosts, rcmd_type, user) < 0)
+                    errx ("%p: Failed to register rcmd \"%s\" for \"%s\"\n",
+                            rcmd_type, hosts);
+            }
         }
     }
 
-    hostlist_destroy (hl);
+    return 0;
+}
 
-    return;
+static int wcoll_args_process (opt_t *opt, char * args)
+{
+    int rc;
+    List l = list_split (",", args);
+    rc = list_for_each (l, (ListForF) wcoll_arg_process, opt);
+    list_destroy (l);
+    return (rc);
+}
+
+static void wcoll_apply_regex (opt_t *opt, List regexs)
+{
+    struct regex_info *re;
+    ListIterator i;
+
+    if (!opt->wcoll || !regexs)
+        return;
+
+    /*
+     *  filter any supplied regular expression args
+     */
+    i = list_iterator_create (regexs);
+    while ((re = list_next (i)))
+        hostlist_filter_regex (opt->wcoll, re);
+    list_iterator_destroy (i);
+}
+
+static void wcoll_apply_excluded (opt_t *opt, List excludes)
+{
+    ListIterator i;
+    char *arg;
+
+    if (!opt->wcoll || !excludes)
+        return;
+
+    /*
+     *  filter explicitly excluded hosts:
+     */
+    i = list_iterator_create (excludes);
+    while ((arg = list_next (i)))
+        hostlist_delete (opt->wcoll, arg);
+    list_iterator_destroy (i);
 }
 
 /*
